import numpy as np
from aiogrpc import insecure_channel
from tensorflow_serving.apis import predict_pb2
from tensorflow_serving.apis import prediction_service_pb2
from tensorflow.contrib.util import make_tensor_proto

class AsyncTensorflowServing(object):
  def __init__(self, endpoint: str):
    self.endpoint = endpoint
    self._channel = None

  async def __aenter__(self):
    self._channel = insecure_channel(self.endpoint)
    self._stub = prediction_service_pb2.beta_create_PredictionService_stub(self._channel)
    return self

  async def __aexit__(self, exc_type, exc, tb):
    await self._channel.close()

  async def predict(self, model_name: str, values,
                    signature_name: str = 'serving_default',
                    input_name: str = 'input',
                    timeout: float = 60.0,
                    dtype: np.dtype = None):
    # We first convert value to a numpy array or scalar.
    if isinstance(values, (np.ndarray, np.generic)):
      if dtype:
        nparray = values.astype(dtype)
      else:
        # python/numpy default float type is float64. We prefer float32 instead.
        if (values.dtype == np.float64) and dtype is None:
          values = values.astype(np.float32)

        # python/numpy default int type is int64. We prefer int32 instead.
        elif (values.dtype == np.int64) and dtype is None:
          downcasted_array = values.astype(np.int32)

          # Do not down cast if it leads to precision loss.
          if np.array_equal(downcasted_array, values):
            values = downcasted_array
    elif callable(getattr(values, "__array__", None)) or isinstance(
        getattr(values, "__array_interface__", None), dict):
      # If a class has the __array__ method, or __array_interface__ dict, then it
      # is possible to convert to numpy array.
      if dtype:
        values = np.asarray(values, dtype=dtype)
      else:
        values = np.asarray(values, dtype=np.float32)
    else:
      if values is None:
        raise ValueError("None values not supported.")

    request = predict_pb2.PredictRequest()
    request.model_spec.name = model_name
    request.model_spec.signature_name = signature_name
    request.inputs[input_name].CopyFrom(make_tensor_proto(values))
    return self._stub.Predict(request, timeout)


__title__ = 'aiotf'
__description__ = 'Asyncio Tensorflow Serving Communication.'
__url__ = 'https://github.com/brunoalano/aiotf'
__version__ = '0.1.0'
__author__ = 'Bruno Alano Medina'
__author_email__ = 'bruno@neurologic.com.br'
__license__ = 'MIT'
__copyright__ = 'Copyright 2018 Bruno Alano Medina'


from .__version__ import __title__, __description__, __url__, __version__
from .__version__ import __author__, __author_email__, __license__
from .__version__ import __copyright__

# Set default logging handler to avoid "No handler found" warnings.
import logging
try:  # Python 2.7+
  from logging import NullHandler
except ImportError:
  class NullHandler(logging.Handler):
    def emit(self, record):
      pass

from .client import AsyncTensorflowServing

__all__ = ['AsyncTensorflowServing']

logging.getLogger(__name__).addHandler(NullHandler())


import os
import sys
from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand

here = os.path.abspath(os.path.dirname(__file__))

class PyTest(TestCommand):
  user_options = [('pytest-args=', 'a', "Arguments to pass into py.test")]

  def initialize_options(self):
    TestCommand.initialize_options(self)
    try:
      from multiprocessing import cpu_count
      self.pytest_args = ['-n', str(cpu_count()), '--boxed']
    except (ImportError, NotImplementedError):
      self.pytest_args = ['-n', '1', '--boxed']

  def finalize_options(self):
    TestCommand.finalize_options(self)
    self.test_args = []
    self.test_suite = True

  def run_tests(self):
    import pytest
    errno = pytest.main(self.pytest_args)
    sys.exit(errno)

with open('README.md') as f:
  long_description = f.read()

py_version = sys.version_info[:2]
if py_version < (3, 4):
  raise Exception("aiotf requires Python >= 3.4.")

about = {}
with open(os.path.join(here, 'aiotf', '__version__.py'), 'r') as f:
  exec(f.read(), about)

setup(
  name=about['__title__'],
  version=about['__version__'],
  description=about['__description__'],
  long_description=long_description,
  long_description_content_type='text/markdown',
  author=about['__author__'],
  author_email=about['__author_email__'],
  url=about['__url__'],
  license=about['__license__'],
  packages=find_packages(exclude=('tests', 'docs')),
  python_requires='>=3.4',
  install_requires=[
    'aiogrpc',
    'tensorflow'
  ],
  cmdclass={'test': PyTest},
  classifiers=(
    "Programming Language :: Python :: 3",
    "License :: OSI Approved :: MIT License",
    "Operating System :: OS Independent",
  ),
)
